from unittest import TestCase

from dimensigon.domain.entities.file import File
from flask_jwt_extended import create_access_token

from dimensigon.domain.entities import Server
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.web import create_app, db


class TestFile(TestCase):
    def setUp(self):
        self.maxDiff = None
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()
        self.src = Server('source', port=5000, id='00000000-0000-0000-0000-000000000000')
        self.dst1 = Server('destination1', port=5000, id='00000000-0000-0000-0000-000000000001')
        self.dst2 = Server('destination2', port=5000, id='00000000-0000-0000-0000-000000000002')
        self.dst3 = Server('destination3', port=5000, id='00000000-0000-0000-0000-000000000003')
        db.session.add_all([self.src, self.dst1, self.dst2, self.dst3])

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_json(self):
        dst_servers = [self.dst1, (self.dst2, '/'), (self.dst3.id, '/')]
        f = File(source_server=self.src, target='/home/dimensigon/dimensigon/dimensigon.log',
                 destination_servers=dst_servers,
                 id='11111111-2222-3333-4444-111111110001')
        db.session.add(f)
        dst_servers = f.destinations

        f_json = f.to_json(destinations=True)

        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server_id='00000000-0000-0000-0000-000000000000',
                 target='/home/dimensigon/dimensigon/dimensigon.log',
                 destinations=[{'dst_server_id': '00000000-0000-0000-0000-000000000001', 'dest_folder': None},
                               {'dst_server_id': '00000000-0000-0000-0000-000000000002', 'dest_folder': '/'},
                               {'dst_server_id': '00000000-0000-0000-0000-000000000003', 'dest_folder': '/'}],
                 dest_folder=None,
                 deleted=False,
                 _old_target=None), f_json)

        f_json = f.to_json()

        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server_id='00000000-0000-0000-0000-000000000000',
                 target='/home/dimensigon/dimensigon/dimensigon.log',
                 dest_folder=None,
                 deleted=False,
                 _old_target=None), f_json)

        f_json = f.to_json(human=True, destinations=True, no_delete=True)

        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 src_server='source',
                 target='/home/dimensigon/dimensigon/dimensigon.log',
                 destinations=[{'dst_server': 'destination1', 'dest_folder': None},
                               {'dst_server': 'destination2', 'dest_folder': '/'},
                               {'dst_server': 'destination3', 'dest_folder': '/'}],
                 dest_folder=None), f_json)

    def test_from_json(self):
        dst_servers = [self.dst1, (self.dst2, '/'), (self.dst3.id, '/')]
        f = File(source_server=self.src, target='/home/dimensigon/dimensigon/dimensigon.log',
                 destination_servers=dst_servers,
                 id='11111111-2222-3333-4444-111111110001')
        dst_servers = f.destinations
        db.session.add(f)
        db.session.commit()
        db.session.close()

        smashed = File.from_json(dict(id='11111111-2222-3333-4444-111111110001',
                                      src_server_id='00000000-0000-0000-0000-000000000000',
                                      target='/home/dimensigon/dimensigon/dimensigon.log',
                                      dest_folder='/home/dimensigon',
                                      deleted=False,
                                      _old_target=None))
        self.assertEqual(smashed.source_server.id, '00000000-0000-0000-0000-000000000000')
        self.assertEqual(smashed.dest_folder, '/home/dimensigon')
        self.assertEqual(smashed.target, '/home/dimensigon/dimensigon/dimensigon.log')
        self.assertFalse(smashed.deleted)
        self.assertListEqual([('00000000-0000-0000-0000-000000000001', None),
                              ('00000000-0000-0000-0000-000000000002', '/'),
                              ('00000000-0000-0000-0000-000000000003', '/')],
                             [(d.destination_server.id, d.dest_folder) for d in smashed.destinations])
        self.assertIsNone(smashed._old_target)

    def test_delete(self):
        dst_servers = [self.dst1, (self.dst2, '/'), (self.dst3.id, '/')]
        f = File(source_server=self.src, target='/home/dimensigon/dimensigon/dimensigon.log',
                 destination_servers=dst_servers,
                 id='11111111-2222-3333-4444-111111110001')

        f.delete()

        self.assertEqual('/home/dimensigon/dimensigon/dimensigon.log', f._old_target)
        self.assertNotEqual('/home/dimensigon/dimensigon/dimensigon.log', f.target)
        self.assertTrue(f.deleted)

        db.session.add(f)
        db.session.commit()

        log_list = f.query.all()
        self.assertEqual(0, len(log_list))
        f = db.session.query(File).filter_by(id='11111111-2222-3333-4444-111111110001').one()
        self.assertTrue(all([d.deleted for d in f.destinations]))




import re
from unittest.case import TestCase
from unittest.mock import patch

from aioresponses import aioresponses
from flask import url_for

from dimensigon.domain.entities import User, Server, Gate, Route
from dimensigon.domain.entities.route import RouteContainer
from dimensigon.use_cases.routing import asyncio, async_update_route_table_cost, update_route_table_from_data
from dimensigon.web import create_app, db


class TestUpdateRouteTableCost(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()
        User.set_initial()
        self.client = self.app.test_client()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario1(self, m, mocked_async_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1')
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3', me=True)
        g3 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                  dns=s3.name)
        Route(s1, g1, cost=0)
        db.session.add_all([s1, s2, s3])

        m.get(re.compile('https?://node1:5001.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000001',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000002',
                                gate_id='00000000-0000-0000-0000-000000000012', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000003',
                                gate_id='00000000-0000-0000-0000-000000000013', proxy_server_id=None, cost=0),
                       ]})
        m.get(re.compile('https?://node2:5002.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000001',
                                gate_id='00000000-0000-0000-0000-000000000011', proxy_server_id=None, cost=0),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True


        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertDictEqual({s2: RouteContainer(None, g2, 0)}
                             , changed_routes)

        self.assertEqual(g1, s1.route.gate)
        self.assertEqual(None, s1.route.proxy_server)
        self.assertEqual(0, s1.route.cost)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertIsNone(s3.route)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario2(self, m, mocked_async_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g3 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                  dns=s3.name)
        Route(s2, g2, cost=0)
        Route(s3, g3, cost=0)
        db.session.add_all([s1, s2, s3])

        m.get(re.compile('https?://node2:5002.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000001',
                                gate_id='00000000-0000-0000-0000-000000000011', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000003',
                                gate_id='00000000-0000-0000-0000-000000000013', proxy_server_id=None, cost=0),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertDictEqual({s3: RouteContainer(s2, None, 1)}, changed_routes)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario3(self, m, mocked_async_check_host):
        # Node 1 loses connection to gate's Node 2 and sets the second gate as default gate
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g21 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5012,
                   dns=s2.name)
        g22 = Gate(id='00000000-0000-0000-0000-000000000022', server=s2, port=5022,
                   dns=s2.name)

        Route(s2, g21, cost=0)
        db.session.add_all([s1, s2])

        m.get(Server.query.get('00000000-0000-0000-0000-000000000002').url('api_1_0.routes').replace(
            '5012', '5022'), payload={"server_id": '00000000-0000-0000-0000-000000000002',
                                      "route_list": [
                                          dict(destination_id='00000000-0000-0000-0000-000000000001',
                                               gate_id='00000000-0000-0000-0000-000000000011', proxy_server_id=None,
                                               cost=0),
                                      ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1' and port == 5001:
                return True
            elif host == 'node2' and port == 5012:
                return False
            elif host == 'node2' and port == 5022:
                return True
            else:
                return False

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g22, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({s2: RouteContainer(
            None, g22, 0)}, changed_routes)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario4(self, m, mocked_async_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g3 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                  dns=s3.name)
        Route(s2, g2, cost=0)
        Route(s3, g3, cost=0)
        db.session.add_all([s1, s2, s3])

        m.get(re.compile('https?://node2:5002.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000001',
                                gate_id='00000000-0000-0000-0000-000000000011', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000003',
                                gate_id=None, proxy_server_id=None, cost=None),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(None, s3.route.cost)

        self.assertDictEqual({s3: RouteContainer(
            None, None, None)}, changed_routes)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario5(self, m, mocked_async_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5012,
                  dns=s2.name)

        Route(s2, g2, cost=0)
        db.session.add_all([s1, s2])

        m.get(re.compile('https?://node2:5002.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000001',
                                gate_id='00000000-0000-0000-0000-000000000011', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000003',
                                gate_id='00000000-0000-0000-0000-000000000013', proxy_server_id=None, cost=0),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            else:
                raise

        mocked_async_check_host.side_effect = async_check_host

        mocked_async_check_host.assert_not_called()

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({}, changed_routes)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario6(self, m, mocked_async_check_host):
        # Nodes have localhost and node2 is not a neighbour anymore
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   ip='127.0.0.1')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s1, port=5000,
                   ip='10.0.0.1')

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g21 = Gate(id='00000000-0000-0000-0000-000000000021', server=s2, port=5000,
                   ip='127.0.0.1')
        g22 = Gate(id='00000000-0000-0000-0000-000000000022', server=s2, port=5000,
                   ip='10.0.0.2')

        Route(s2, g22, cost=0)
        db.session.add_all([s1, s2])
        db.session.commit()

        m.get(re.compile('https?://node1:5001.*'),
              payload={"server_id": '00000000-0000-0000-0000-000000000001',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000002',
                                gate_id='00000000-0000-0000-0000-000000000022', proxy_server_id=None, cost=0),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == '127.0.0.1':
                return True
            elif host == '10.0.0.1':
                return True
            elif host == '10.0.0.2':
                return False
            else:
                raise ConnectionError

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(None, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(None, s2.route.cost)

        self.assertDictEqual({s2: RouteContainer(
            None, None, None)}, changed_routes)

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario6(self, m, mocked_async_check_host):
        # Node have localhost and node2 appears as a new neighbour
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   ip='127.0.0.1')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s1, port=5000,
                   ip='10.0.0.1')

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g21 = Gate(id='00000000-0000-0000-0000-000000000021', server=s2, port=5000,
                   ip='127.0.0.1')
        g22 = Gate(id='00000000-0000-0000-0000-000000000022', server=s2, port=5000,
                   ip='10.0.0.2')

        Route(s2, None, cost=None)
        db.session.add_all([s1, s2])
        db.session.commit()

        m.get(re.compile("^https?://(10\.0\.0\.1|127\.0\.0\.1):5000" + url_for('api_1_0.routes', _external=False)),
              payload={"server_id": '00000000-0000-0000-0000-000000000001',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000002',
                                gate_id='00000000-0000-0000-0000-000000000022', proxy_server_id=None, cost=0),
                       ]})

        m.get(re.compile("^https?://(10\.0\.0\.2):5000" + url_for('api_1_0.routes', _external=False)),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000001',
                                gate_id='00000000-0000-0000-0000-000000000012', proxy_server_id=None, cost=0),
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == '127.0.0.1':
                return True
            elif host == '10.0.0.1':
                return True
            elif host == '10.0.0.2':
                return True
            else:
                raise ConnectionError

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g22, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({s2: RouteContainer(
            None, g22, 0)}, changed_routes)

        db.session.commit()

    @aioresponses()
    @patch('dimensigon.use_cases.routing.async_check_host', autospec=True)
    def test_update_table_routing_cost_scenario6(self, m, mocked_async_check_host):
        # Node have localhost and node2 appears as a new neighbour
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g21 = Gate(id='00000000-0000-0000-0000-000000000021', server=s2)

        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g31 = Gate(id='00000000-0000-0000-0000-000000000031', server=s3)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g41 = Gate(id='00000000-0000-0000-0000-000000000041', server=s4)

        Route(s2, g21, cost=0)
        Route(s3, None, cost=None)
        Route(s4, None, cost=None)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        m.get(re.compile("^https?://node1:\d+" + url_for('api_1_0.routes', _external=False)),
              payload={"server_id": '00000000-0000-0000-0000-000000000001',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000002',
                                gate_id='00000000-0000-0000-0000-000000000022', proxy_server_id=None, cost=0),
                       ]})

        m.get(re.compile("^https?://node2:\d+" + url_for('api_1_0.routes', _external=False)),
              payload={"server_id": '00000000-0000-0000-0000-000000000002',
                       "route_list": [
                           dict(destination_id='00000000-0000-0000-0000-000000000002',
                                gate_id='00000000-0000-0000-0000-000000000021', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000003',
                                gate_id='00000000-0000-0000-0000-000000000031', proxy_server_id=None, cost=0),
                           dict(destination_id='00000000-0000-0000-0000-000000000004',
                                gate_id=None, proxy_server_id='00000000-0000-0000-0000-000000000003', cost=1)
                       ]})

        async def async_check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            elif host == 'node4':
                return False
            else:
                raise ConnectionError

        mocked_async_check_host.side_effect = async_check_host

        changed_routes = asyncio.run(
            async_update_route_table_cost(discover_new_neighbours=True, check_current_neighbours=True))

        self.assertIsNone(s1.route)

        self.assertEqual(g21, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertEqual(None, s4.route.gate)
        self.assertEqual(s2, s4.route.proxy_server)
        self.assertEqual(2, s4.route.cost)

        self.assertDictEqual({s3: RouteContainer(s2, None, 1),
                              s4: RouteContainer(s2, None, 2),
                              }, changed_routes)


class TestUpdateRouteTableFromData(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.maxDiff = None
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()
        User.set_initial()
        self.client = self.app.test_client()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch(self, mocked_ping):

        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='server1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='server2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                  dns=s2.name)
        Route(s2, g2, cost=0)

        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='server3')
        g3 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                  dns=s3.name)
        Route(s3, g3, cost=0)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='server4')
        g4 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5001,
                  dns=s4.name)
        Route(s4, s2, cost=1)
        db.session.add_all([s1, s2, s3, s4])

        # Server2 loses connectivity to Server4
        mocked_ping.return_value = (None, None)

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000004',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None)]})

        self.assertDictEqual({s4: RouteContainer(None, None, None)}, new_routes)

        # s = Server.query.get('00000000-0000-0000-0000-000000000002')
        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        # s = Server.query.get('00000000-0000-0000-0000-000000000003')
        self.assertEqual(g3, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

        # s = Server.query.get('00000000-0000-0000-0000-000000000004')
        self.assertEqual(None, s4.route.gate)
        self.assertEqual(None, s4.route.proxy_server)
        self.assertEqual(None, s4.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario01(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g1 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                  dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g2 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g3 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                  dns=s3.name)

        Route(s2, g2, cost=0)
        db.session.add_all([s1, s2, s3])
        db.session.commit()


        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000003',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000001',
                                                            gate_id='00000000-0000-0000-0000-000000000011',
                                                            proxy_server_id=None,
                                                            cost=0),
                                                       dict(destination_id='00000000-0000-0000-0000-000000000002',
                                                            gate_id='00000000-0000-0000-0000-000000000012',
                                                            proxy_server_id=None,
                                                            cost=0)
                                                   ]})

        self.assertDictEqual({s3: RouteContainer(None, g3, 0)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(g3, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario02(self,
                                     mocked_ping,
                                     mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        Route(s2, g12, cost=0)
        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id='00000000-0000-0000-0000-000000000013',
                                                            proxy_server_id=None,
                                                            cost=0),
                                                       dict(destination_id='00000000-0000-0000-0000-000000000002',
                                                            gate_id='00000000-0000-0000-0000-000000000012',
                                                            proxy_server_id=None,
                                                            cost=0)
                                                   ]})

        self.assertDictEqual({s3: RouteContainer(s2, None, 1)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario03(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        Route(s2, g12, cost=0)
        Route(s3, s2, cost=1)
        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000003',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000001',
                                                            gate_id='00000000-0000-0000-0000-000000000011',
                                                            proxy_server_id=None,
                                                            cost=0)
                                                   ]})

        self.assertDictEqual({s3: RouteContainer(None, g13, 0)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(g13, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario04(self,
                                     mocked_ping,
                                     mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000001',
                                                            gate_id='00000000-0000-0000-0000-000000000011',
                                                            proxy_server_id=None,
                                                            cost=0),
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id='00000000-0000-0000-0000-000000000013',
                                                            proxy_server_id=None,
                                                            cost=0)
                                                   ]})
        self.assertDictEqual({s2: RouteContainer(None, g12, 0),
                              s3: RouteContainer(s2, None, 1)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario05(self,
                                     mocked_ping,
                                     mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        Route(s2, g12, cost=0)
        Route(s3, g13, cost=0)
        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None)
                                                   ]})

        self.assertDictEqual({}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(g13, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario06(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        Route(s2, g12, cost=0)
        Route(s3, g13, cost=0)
        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": s2.id,
                                                   "route_list": [
                                                       dict(destination_id=s3.id,
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None)
                                                   ]})

        self.assertDictEqual({s3: RouteContainer(None, None, None)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(None, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario07(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5001,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5002,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5003,
                   dns=s3.name)

        Route(s2, g12, cost=0)
        Route(s3, s2, cost=1)
        db.session.add_all([s1, s2, s3])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None)
                                                   ]})

        self.assertDictEqual({s3: RouteContainer(None, None, None)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(None, s3.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario08(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000,
                   dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000,
                   dns=s4.name)

        Route(s2, g12, cost=0)
        Route(s3, s2, cost=1)
        Route(s4, s2, cost=1)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            elif host == 'node4':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000004',
                                                            gate_id=None,
                                                            proxy_server_id=s3.id,
                                                            cost=1),

                                                   ]})

        self.assertDictEqual({s4: RouteContainer(s2, None, 2)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertEqual(None, s4.route.gate)
        self.assertEqual(s2, s4.route.proxy_server)
        self.assertEqual(2, s4.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario09(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000, dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000, dns=s2.name)

        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000, dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000, dns=s4.name)

        Route(s2, g12, cost=0)
        Route(s3, s2, cost=1)
        Route(s4, s2, cost=2)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            elif host == 'node4':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000004',
                                                            proxy_server_id=None,
                                                            gate_id=g14.id,
                                                            cost=0)
                                                   ]})

        self.assertDictEqual({s4: RouteContainer(s2, None, 1)}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertEqual(None, s4.route.gate)
        self.assertEqual(s2, s4.route.proxy_server)
        self.assertEqual(1, s4.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario10(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000, dns=s1.name)

        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000, dns=s2.name)

        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000, dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000, dns=s4.name)

        s5 = Server(id='00000000-0000-0000-0000-000000000005', name='node5')
        g15 = Gate(id='00000000-0000-0000-0000-000000000015', server=s5, port=5000, dns=s5.name)

        Route(s2, g12)
        Route(s3, g13)
        Route(s4, s3, 1)
        Route(s5, s2, 1)
        db.session.add_all([s1, s2, s3, s4, s5])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True
            elif host == 'node4':
                return False
            elif host == 'node5':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        def ping(server, *args, **kwargs):
            if str(server.id) == '00000000-0000-0000-0000-000000000001':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000002':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000003':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000004':
                return 1, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000005':
                return 1, None
            else:
                raise

        mocked_ping.side_effect = ping

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000003',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000005',
                                                            proxy_server_id=s4.id,
                                                            gate_id=None,
                                                            cost=1)
                                                   ]})

        self.assertDictEqual({}, new_routes)

        self.assertEqual(0, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(g13, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

        self.assertEqual(None, s4.route.gate)
        self.assertEqual(s3, s4.route.proxy_server)
        self.assertEqual(1, s4.route.cost)

        self.assertEqual(None, s5.route.gate)
        self.assertEqual(s2, s5.route.proxy_server)
        self.assertEqual(1, s5.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario11(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000,
                   dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000,
                   dns=s4.name)

        Route(s2, g12)
        Route(s3, s4, 1)
        Route(s4, g14)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            elif host == 'node4':
                return True
            else:
                raise

        mocked_check_host.side_effect = check_host

        def ping(server, *args, **kwargs):
            if str(server.id) == '00000000-0000-0000-0000-000000000001':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000002':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000003':
                return 1, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000004':
                return 0, None
            else:
                raise

        mocked_ping.side_effect = ping

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None),

                                                   ]})

        self.assertDictEqual({}, new_routes)

        self.assertEqual(1, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s4, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertEqual(g14, s4.route.gate)
        self.assertEqual(None, s4.route.proxy_server)
        self.assertEqual(0, s4.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario12(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000,
                   dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000,
                   dns=s4.name)

        Route(s2, g12)
        Route(s3, s4, 1)
        Route(s4, g14)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False
            elif host == 'node4':
                return True
            else:
                raise

        mocked_check_host.side_effect = check_host

        def ping(server, *args, **kwargs):
            if str(server.id) == '00000000-0000-0000-0000-000000000001':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000002':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000003':
                return None, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000004':
                return 0, None
            else:
                raise

        mocked_ping.side_effect = ping

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000002',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000003',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None),

                                                   ]})

        self.assertDictEqual({s3: RouteContainer(None, None, None)}, new_routes)

        self.assertEqual(1, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(g12, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(None, s3.route.cost)

        self.assertEqual(g14, s4.route.gate)
        self.assertEqual(None, s4.route.proxy_server)
        self.assertEqual(0, s4.route.cost)

    @patch('dimensigon.use_cases.routing.check_host')
    @patch('dimensigon.use_cases.routing.ntwrk.ping')
    def test_routes_patch_scenario13(self, mocked_ping, mocked_check_host):
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', me=True)
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000,
                   dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000,
                   dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000,
                   dns=s3.name)

        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000,
                   dns=s4.name)

        Route(s2, s3, 1)
        Route(s3, g13)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return False
            elif host == 'node3':
                return False
            elif host == 'node4':
                return False
            else:
                raise

        mocked_check_host.side_effect = check_host

        def ping(server, *args, **kwargs):
            if str(server.id) == '00000000-0000-0000-0000-000000000001':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000002':
                return 1, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000003':
                return 0, None
            elif str(server.id) == '00000000-0000-0000-0000-000000000004':
                return None, None
            else:
                raise

        mocked_ping.side_effect = ping

        new_routes = update_route_table_from_data({"server_id": '00000000-0000-0000-0000-000000000004',
                                                   "route_list": [
                                                       dict(destination_id='00000000-0000-0000-0000-000000000002',
                                                            gate_id=None,
                                                            proxy_server_id=None,
                                                            cost=None),

                                                   ]})

        self.assertDictEqual({}, new_routes)

        self.assertEqual(1, mocked_ping.call_count)
        self.assertEqual(0, mocked_check_host.call_count)
        self.assertIsNone(s1.route)

        self.assertEqual(None, s2.route.gate)
        self.assertEqual(s3, s2.route.proxy_server)
        self.assertEqual(1, s2.route.cost)

        self.assertEqual(g13, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(0, s3.route.cost)

        self.assertEqual(None, s4.route.gate)
        self.assertEqual(None, s4.route.proxy_server)
        self.assertEqual(None, s4.route.cost)



import datetime as dt
import json
import logging
import queue
import typing as t

from flask import current_app

from dimensigon.domain.entities import Server
from dimensigon.use_cases import routing
from dimensigon.use_cases.helpers import get_root_auth
from dimensigon.utils import asyncio
from dimensigon.utils.cluster import Cluster
from dimensigon.utils.helpers import format_exception
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, threading

logger = logging.getLogger('dimensigon.cluster')


async def send_cluster_register(cr, servers=None, auth=None, exclude=None):
    if not servers:
        servers = Server.get_neighbours(exclude=exclude)
    responses = await ntwrk.parallel_requests(servers, 'post',
                                              view_or_url='api_1_0.cluster',
                                              json=cr,
                                              auth=auth or get_root_auth(),
                                              timeout=10)

    for r, s in zip(responses, servers):
        if not r.ok and logger.level <= logging.WARNING:
            logger.warning(
                f"Unable to send cluster information to {s}. Response: {r}")


def check_server_alive(server: Server):
    alive_server_ids = [i for i in current_app.cluster_manager.cluster if
                        server.id != i and i != Server.get_current().id]
    # check if I have it as a neighbour
    if server.route and server.route.cost == 0:
        route = routing.check_gates(server)
        if route:
            return True

    # in order to prevent broadcast to everyone, first try a ping
    cost, elapsed = ntwrk.ping(server, retries=1, timeout=15)
    if cost is not None:
        return True
    responses = asyncio.run(ntwrk.parallel_requests(alive_server_ids, 'get',
                                                    view_or_url='api_1_0.routes_neighbour',
                                                    view_data=dict(server_id=server.id), timeout=10))
    for r in responses:
        if r.ok:
            if r.msg.get('neighbour'):
                return True
    return False


def update_cluster_status():
    alive_server_ids = current_app.cluster_manager.cluster.get_alive()
    updated = False
    for alive_server_id in alive_server_ids:
        if alive_server_id != Server.get_current().id:
            alive = check_server_alive(Server.query.get(Server))
            if not alive:
                current_app.cluster_manager.cluster.set_death(alive_server_id)
                updated = True
    return updated


async def check_heartbeat_and_send(cluster_session_id: int, heartbeat_id: dt.datetime,
                                   exclude: t.Optional[t.List[Id]] = None):
    cr = current_app.cluster_manager.cluster.set_alive(cluster_session_id, heartbeat_id)
    if cr:
        await send_cluster_register(cr, exclude=exclude)


class ClusterManager:

    def __init__(self, app, maxsize=None, threshold=None, retain_time=2, start=True):
        self.app = app
        self.queue = queue.Queue(maxsize=maxsize or 10000)
        self.cluster = Cluster(threshold=threshold)
        self._stop = False
        self.buffer: t.Dict[Id, t.Dict] = {}
        self.retain_time = retain_time
        self.loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self.run, name='cluster_manager')
        self._change_buffer_lock = threading.RLock()
        self._timer = None
        if start:
            self._thread.start()

    @property
    def running(self):
        return not self._stop

    def start(self):
        self._thread.start()

    def stop(self):
        self._stop = True
        self.queue.put('STOP', block=False)
        if self._timer:
            self._timer.cancel()

    def _add2buffer(self, data: t.Union[t.List[t.Dict], t.Dict]):
        with self._change_buffer_lock:
            if isinstance(data, dict):
                self.buffer.update({data['id']: data})
            else:
                self.buffer.update(**{cr['id']: cr for cr in data})
            if self._timer is None:
                self._timer = threading.Timer(interval=self.retain_time, function=self.send_data)
                self._timer.start()

    def set_alive(self, ident, alive=None):
        cr = self.cluster.set_alive(ident, alive)
        if cr:
            with self._change_buffer_lock:
                self._add2buffer(cr)

    def set_death(self, ident, death=None):
        cr = self.cluster.set_death(ident, death)
        if cr:
            with self._change_buffer_lock:
                self._add2buffer(cr)

    def set_keepalive(self, ident: Id, keepalive=None):
        cr = self.cluster.set_keepalive(ident, keepalive)
        if cr:
            with self._change_buffer_lock:
                self._add2buffer(cr)

    def put(self, data: t.Union[t.Dict, t.List[t.Dict]], block=True, timeout=None):
        self.queue.put(data, block=block, timeout=timeout)

    def send_data(self):
        with self.app.app_context():
            # time to send data
            neighbours = Server.get_neighbours()
            if neighbours:
                auth = get_root_auth()
                with self._change_buffer_lock:
                    temp_buffer = dict(self.buffer)
                    self.buffer.clear()

                logger.debug(
                    f"Sending cluster information to the following nodes {', '.join([s.name for s in neighbours])}: "
                    f"{json.dumps(log_data(temp_buffer.values()), indent=2)}")
                try:
                    responses = self.loop.run_until_complete(
                        ntwrk.parallel_requests(neighbours, 'POST', view_or_url='api_1_0.cluster',
                                                json=list(temp_buffer.values()), auth=auth))
                except Exception as e:
                    logger.error(f"Unable to send cluster information to neighbours: "
                                 f"{format_exception(e)}")
                    # restore data with new data arrived
                    with self._change_buffer_lock:
                        temp_buffer.update(**self.buffer)
                        self.buffer.clear()
                        self.buffer.update(temp_buffer)
                else:
                    for s, r in zip(neighbours, responses):
                        if not r.ok:
                            logger.warning(f"Unable to send data to {s}: {r}")

                # check if new data arrived during timer execution
                with self._change_buffer_lock:
                    if self.buffer:
                        self._timer = threading.Timer(interval=1, function=self.send_data)
                        self._timer.start()
                    else:
                        self._timer = None
            else:
                logger.debug(f"No neighbour servers to send cluster information")
                with self._change_buffer_lock:
                    self._timer = None

    def run(self):
        logger.debug('Starting cluster manager')
        while not self._stop:
            try:
                item = self.queue.get(block=True, timeout=5)
            except queue.Empty:
                pass
            else:
                if item == 'STOP':
                    break
                elif item:
                    changed_ids = self.cluster.update_cluster(item)
                    for c_id in changed_ids:
                        self._add2buffer(self.cluster.get(c_id))
        logger.debug('Stopping cluster manager')


def log_data(data):
    debug_data = []
    for cr in data:
        cr = dict(cr)
        cr['name'] = getattr(Server.query.get(cr['id']), 'name', cr['id'])
        debug_data.append(cr)
    return debug_data



import datetime as dt
import random
from unittest import TestCase, mock

from dimensigon import defaults
from dimensigon.domain.entities import Server, Route, Scope
from dimensigon.use_cases.helpers import get_servers_from_scope
from dimensigon.web import create_app, db

old_age = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
now = dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc)


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dimensigon.use_cases.helpers.current_app.cluster.get_delta_keepalive')
    @mock.patch('dimensigon.use_cases.helpers.get_now')
    def test_get_servers_from_scope_one_server(self, mock_get_now, mock_keepalive):
        mock_get_now.return_value = now
        s0 = Server(id='00000000-0000-0000-0000-000000000000', name='node0', port=5000, me=True, created_on=old_age)
        db.session.add(s0)
        mock_keepalive.return_value = [s0.id]

        quorum = get_servers_from_scope(scope=Scope.CATALOG)
        self.assertEqual([s0], quorum)

        s0.created_on = now

        quorum = get_servers_from_scope(scope=Scope.CATALOG)
        self.assertEqual([s0], quorum)

        mock_keepalive.return_value = []

        quorum = get_servers_from_scope(scope=Scope.CATALOG)
        self.assertEqual([s0], quorum)

        s0.l_ignore_on_lock = True

        quorum = get_servers_from_scope(scope=Scope.CATALOG)
        self.assertEqual([], quorum)

    @mock.patch('dimensigon.use_cases.helpers.defaults.MIN_SERVERS_QUORUM', 2)
    @mock.patch('dimensigon.use_cases.helpers.current_app.cluster.get_delta_keepalive')
    @mock.patch('dimensigon.use_cases.helpers.get_now')
    def test_get_servers_from_scope_young_server(self, mock_get_now, mock_keepalive):
        mock_get_now.return_value = now
        s0 = Server(id='00000000-0000-0000-0000-000000000000', name='node0', port=5000, me=True, created_on=now)
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', port=5000, created_on=now)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2', port=5000, created_on=now)
        db.session.add_all([s0, s1, s2])
        mock_keepalive.return_value = [s1.id, s2.id]

        quorum = get_servers_from_scope(scope=Scope.CATALOG)
        self.assertEqual([s0, s1], quorum)

    @mock.patch('dimensigon.use_cases.helpers.defaults.MIN_SERVERS_QUORUM', 5)
    @mock.patch('dimensigon.use_cases.helpers.current_app.cluster.get_delta_keepalive')
    @mock.patch('dimensigon.domain.entities.get_now')
    def test_get_servers_from_scope_more_than_min_quorum(self, mock_get_now, mock_keepalive):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

        Server.set_initial()
        servers = []
        for i in range(0, 7):

            s = Server(f'node{i}', port=5000, created_on=old_age)

            if i == 0:
                r = Route(s, cost=0)
            else:
                r = Route(s, random.choice(servers), cost=i)

            db.session.add_all([s, r])

            servers.append(s)

        mock_get_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        mock_keepalive.return_value = [s.id for s in servers]
        s62 = Server(f'node72', port=5000)
        Route(s62, random.choice(servers), cost=6)
        db.session.add(s62)

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(8, len(quorum))
        self.assertNotIn(s62, quorum)
        self.assertIn(s, quorum)
        self.assertIn(Server.get_current(), quorum)

    @mock.patch('dimensigon.use_cases.helpers.defaults.MIN_SERVERS_QUORUM', 5)
    @mock.patch('dimensigon.use_cases.helpers.current_app.cluster.get_delta_keepalive')
    @mock.patch('dimensigon.domain.entities.get_now')
    def test_get_servers_from_scope_less_than_min_quorum(self, mock_get_now, mock_keepalive):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        me = Server(id='00000000-0000-0000-0000-000000000000', name='node0', port=5000, me=True, created_on=now)
        db.session.add(me)
        servers = []
        for i in range(1, 4):
            s = Server(f'node{i}', port=5000, created_on=old_age)
            r = Route(s, cost=0)

            db.session.add_all([s, r])

            servers.append(s)

        mock_keepalive.return_value = [s.id for s in servers]

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(4, len(quorum))
        self.assertIn(Server.get_current(), quorum)

    @mock.patch('dimensigon.use_cases.helpers.defaults.MIN_SERVERS_QUORUM', 5)
    @mock.patch('dimensigon.use_cases.helpers.current_app.cluster.get_delta_keepalive')
    @mock.patch('dimensigon.domain.entities.get_now')
    def test_get_servers_from_scope_more_than_min_quorum_no_cost(self, mock_get_now, mock_keepalive):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        Server.set_initial()
        servers = []
        for i in range(0, 7):

            s = Server(f'node{i}', port=5000, created_on=old_age)

            if i == 0:
                r = Route(s, cost=0)
            else:
                r = Route(s, cost=0)

            db.session.add_all([s, r])

            servers.append(s)
        mock_keepalive.return_value = [s.id for s in servers]

        mock_get_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        s62 = Server(f'node72', port=5000)
        Route(s62, proxy_server=random.choice(servers), cost=6)
        db.session.add(s62)

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(defaults.MIN_SERVERS_QUORUM+1, len(quorum))
        self.assertIn(Server.get_current(), quorum)



import configparser
import logging
import os
import re
import urllib

import requests
from prompt_toolkit import prompt

from dimensigon import defaults
from dimensigon.dshell import environ as env
from dimensigon.dshell.view_path_mapping import view_path_map
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.utils.helpers import is_iterable_not_string
from dimensigon.web.network import Response

logger = logging.getLogger('dshell.network')




def exists_refresh_token():
    return bool(env._refresh_token)


def bootstrap_auth(username=None, password=None, refresh_token=None):

    if username:
        env._username = username
    if refresh_token:
        env._refresh_token = refresh_token

    if not exists_refresh_token():
        login(username=username, password=password)
    else:
        try:
            refresh_username = refresh_access_token()
        except Exception:
            raise
        else:
            if refresh_username:
                if env._username is None:
                    env._username = refresh_username
                elif env._username != refresh_username:
                    raise ValueError(f"user '{env._username}' does not correspond with the user from the token")


def login(username=None, password=None):

    if not username:
        if not env._username:
            try:
                username = prompt("Username: ")
            except KeyboardInterrupt:
                return
        else:
            username = env._username
    if not password:
        try:
            password = prompt("Password: ", is_password=True)
        except KeyboardInterrupt:
            return

    resp = requests.post(generate_url('root.login', {}), json={'username': username, 'password': password},
                         verify=False)
    resp.raise_for_status()

    env._username = username
    env._access_token = resp.json()['access_token']
    env._refresh_token = resp.json()['refresh_token']

    file = os.path.expanduser(env.get('CONFIG_FILE', None))
    if file and os.path.exists(file):
        config = configparser.ConfigParser()
        config.read(file)
        if config.has_section('AUTH') and config.has_section('REMOTE'):
            if config['AUTH'].get('username') == env._username:
                from dimensigon.dshell.bootstrap import save_config_file
                del config
                save_config_file(token=env._refresh_token)


def refresh_access_token(login_=True):
    url = f"{env.get('SCHEME')}://{env.get('SERVER')}:{env.get('PORT')}/refresh"

    if env._refresh_token is None:
        raise ValueError('empty refresh token. Login first')
    auth = HTTPBearerAuth(env._refresh_token)
    resp = requests.post(url, auth=auth, verify=False, timeout=10)
    if resp.status_code in (401, 422) and login_:
        login()
    else:
        resp.raise_for_status()
        env._access_token = resp.json().get('access_token', None)
        return resp.json().get('username', None)


def request(method, url, session=None, token_refreshed=False, login=True, **kwargs) -> Response:
    exception = None
    content = None
    status = None
    json_data = None
    headers = {}

    if not session:
        _session = requests.session()
    else:
        _session = session

    raise_on_error = kwargs.pop('raise_on_error', False)

    func = getattr(_session, method.lower())

    if 'auth' not in kwargs:
        if env._access_token is None:
            try:
                refresh_access_token(login_=login)
            except requests.exceptions.ConnectionError as e:
                return Response(exception=ConnectionError(f"Unable to contact with {env.get('SCHEME')}://"
                                                     f"{env.get('SERVER')}:{env.get('PORT')}/refresh"),
                                url=url)
            except Exception as e:
                return Response(exception=e, url=url)
            else:
                if env._access_token is None:
                    return Response(exception=ValueError("No authentication set"), url=url)
        kwargs['auth'] = HTTPBearerAuth(env._access_token)

    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    kwargs['headers'].update({'D-Securizer': 'plain'})

    kwargs['verify'] = False

    logger.debug(f"{method.upper()} {url}\n{kwargs}")

    resp = None
    try:
        resp: requests.Response = func(url, **kwargs)
    except (requests.Timeout,) as e:
        timeout = kwargs.get('timeout', None)
        if isinstance(timeout, tuple):
            timeout = timeout[0] + timeout[1]
        exception = TimeoutError(f"Socket timeout reached while trying to connect to {url} "
                                 f"for {timeout} seconds")
    except requests.ConnectionError as e:
        exception = ConnectionError(f"Unable to contact to {url}")
    except Exception as e:
        exception = e
    finally:
        if session is None and not (
                getattr(resp, 'status_code', None) == 401 and resp.json().get('msg') == 'Token has expired'):
            _session.close()
    if exception is None:
        status = resp.status_code
        headers = resp.headers
        if status == 401 and not token_refreshed:
            json_data = resp.json()
            if json_data.get('msg') == 'Token has expired':
                try:
                    refresh_access_token()
                except requests.exceptions.ConnectionError as e:
                    return Response(exception=ConnectionError(f"Unable to contact with {env.get('SCHEME')}://"
                                                         f"{env.get('SERVER')}:{env.get('PORT')}/refresh"),
                                    url=url)
                kwargs['auth'] = HTTPBearerAuth(env._access_token)
                resp = request(method, url, session=_session, token_refreshed=True, **kwargs)
                if not session:
                    _session.close()
                return resp
        try:
            json_data = resp.json()
        except (ValueError,):
            content = resp.text

        if json_data is not None:
            # try:
            #     content = unpack_msg(json_data)
            # except NotValidMessage:
            #     content = json_data
            content = json_data
    else:
        if raise_on_error:
            raise exception

    return Response(msg=content, code=status, exception=exception, url=url, headers=headers)


def get_parameters_from_path(view):
    path = view_path_map[view]
    match_iterator = re.finditer('\<([\w_]+)\>', path)
    for match in match_iterator:
        yield match.groups()[0]


def _replace_path_args(path, args):
    match_iterator = re.finditer(r'\<((\w+:)?([\w_]+))\>', path)
    replaced_path = path
    for match in match_iterator:
        text = match.groups()[2]
        assert text in args
        value = args.pop(text)
        if not value:
            raise ValueError(f"No value specified for '{text}' in URL {path}")
        replaced_path = "{}{}{}".format(replaced_path[:match.start()], value, replaced_path[match.end():])

    params = []
    if args:
        for k, v in args.items():
            if is_iterable_not_string(v):
                for vv in v:
                    if vv is not None:
                        params.append(f"{k}={urllib.parse.quote_plus(vv)}")
            else:
                if v is not None:
                    params.append(f"{k}={urllib.parse.quote_plus(v)}")
    return replaced_path + '?' + '&'.join(params)


def generate_url(view, view_data, ip=None, port=defaults.DEFAULT_PORT, scheme='https'):
    path = view_path_map[view]
    path = _replace_path_args(path, view_data or {})
    if (env.get('SERVER', ip) or ip) is None:
        raise ValueError('No SERVER specified.')
    return f"{env.get('SCHEME', scheme) or scheme}://" \
           f"{env.get('SERVER', ip) or ip}:" \
           f"{env.get('PORT', port) or port}" \
           f"{path}"


def get(view, view_data=None, **kwargs) -> Response:
    return request('get', generate_url(view, view_data), **kwargs)


def post(view, view_data=None, **kwargs) -> Response:
    return request('post', generate_url(view, view_data), **kwargs)


def put(view, view_data=None, **kwargs) -> Response:
    return request('put', generate_url(view, view_data), **kwargs)


def patch(view, view_data=None, **kwargs) -> Response:
    return request('patch', generate_url(view, view_data), **kwargs)


def delete(view, view_data=None, **kwargs) -> Response:
    return request('delete', generate_url(view, view_data), **kwargs)



import datetime as dt
from unittest import TestCase
from unittest.mock import patch

import responses
from aioresponses import aioresponses
from flask_jwt_extended import create_access_token

from dimensigon.domain.entities import Locker, Catalog
from dimensigon.domain.entities import Server, Route, Dimension, User
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.utils.helpers import generate_dimension
from dimensigon.web import create_app, db
from dimensigon.web.background_tasks import process_catalog_route_table
from tests.helpers import set_callbacks


class TestProcessCatalogRouteTable(TestCase):

    @patch('dimensigon.domain.entities.get_now')
    def setUp(self, mock_now):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app.config['SECURIZER'] = True
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            self.dim = generate_dimension('test')
            self.dim.current = True
            self.json_dim = self.dim.to_json()
            self.client = self.app.test_client()

            db.create_all()
            Locker.set_initial()
            User.set_initial()

            self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

            server = Server('node1', dns_or_ip='127.0.0.1', port=8000, me=True)
            for g in server.gates:
                g.last_modified_at = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
            db.session.add_all([server, self.dim])
            db.session.commit()

            # dump data
            self.json_node1 = server.to_json(add_gates=True)
            self.json_users = [u.to_json() for u in User.query.all()]

        self.app2 = create_app('test')
        self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()
        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        with self.app2.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()

            me = Server('node2', port=8000, me=True, granules='granule',
                        last_modified_at=dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc))
            for g in me.gates:
                g.last_modified_at = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
            db.session.add(me)

            src_server = Server.from_json(self.json_node1)
            Route(src_server, cost=0)
            db.session.add(src_server)

            dim = Dimension.from_json(self.json_dim)
            dim.current = True
            db.session.add(dim)

            users = [User.from_json(ju) for ju in self.json_users]
            db.session.add_all(users)

            db.session.commit()

            # dump data
            self.json_node2 = me.to_json(add_gates=True)

        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            node2 = Server.from_json(self.json_node2)
            node2.last_modified_at = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
            Route(node2, cost=0)
            db.session.add(node2)
            db.session.commit()

    def tearDown(self) -> None:
        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

        with self.app.app_context():
            db.session.remove()
            db.drop_all()


    @responses.activate
    @aioresponses()
    @patch('dimensigon.web.background_tasks.update_route_table_cost', return_value=True)
    @patch('dimensigon.web.background_tasks.upgrade_version', return_value=False)
    def test_catalog(self, m, mock_version, mock_routing):
        # test all system from process_catalog_route_table to lock server and upgrade catalog
        set_callbacks([("(127.0.0.1|node1)", self.app.test_client()),
                       ("node2", self.app2.test_client())], m=m)

        with self.app.app_context():
            datemark = Catalog.max_catalog()
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), datemark)

        process_catalog_route_table(self.app)

        with self.app.app_context():
            datemark = Catalog.max_catalog()
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), datemark.astimezone(dt.timezone.utc))





import base64
import functools
import ipaddress
import json
import logging
import socket
import time
import typing as t

import requests
import rsa
from flask import current_app, url_for, g, request
from flask_jwt_extended import get_jwt_claims
from jsonschema import validate, SchemaError

from dimensigon import defaults
from dimensigon.domain.entities import Server, Scope, User, Locker, State, Gate
from dimensigon.network.exceptions import NotValidMessage
from dimensigon.use_cases.helpers import get_servers_from_scope
from dimensigon.use_cases.lock import lock_scope
from dimensigon.utils.helpers import get_now
from dimensigon.web import db, errors
from dimensigon.web.network import unpack_msg, pack_msg, unpack_msg2, pack_msg2

if t.TYPE_CHECKING:
    import flask


def save_if_hidden_ip(remote_addr: str, server: Server):
    ip = ipaddress.ip_address(remote_addr)
    if not ip.is_loopback:
        def get_ip(dns: str):
            if dns is None:
                return
            else:
                try:
                    return ipaddress.ip_address(socket.gethostbyname(dns))
                except:
                    return

        gate = [gate for gate in server.gates if ip in (gate.ip, get_ip(gate.dns))]
        if not gate:
            try:
                with lock_scope(Scope.CATALOG):
                    for port in set([gate.port for gate in server.gates]):
                        gate = server.add_new_gate(dns_or_ip=remote_addr, port=port, hidden=True)
                        db.session.add(gate)
                    db.session.commit()
            except errors.LockerError:
                logger = logging.getLogger('dimensigon')
                logger.warning(f"unable to save {remote_addr} from {server}")


def _proxy_request(request: 'flask.Request', destination: Server, verify=False) -> requests.Response:

    url = destination.url() + request.full_path
    req_data = request.get_json()

    if request.path == '/ping':
        server_data = {'id': str(g.server.id), 'name': g.server.name,
                       'time': get_now().strftime(defaults.DATETIME_FORMAT)}
        if req_data:
            if 'servers' not in req_data:
                req_data['servers'] = {}
            req_data['servers'].update({len(req_data['servers']) + 1: server_data})
        else:
            req_data = dict(servers={1: server_data})
    kwargs = {
        'json': req_data,
        'allow_redirects': False
    }

    headers = {key.lower(): value for key, value in request.headers.items()}

    # Let requests reset the host for us.
    if 'host' in headers:
        del headers['host']

    headers['d-source'] = headers.get('d-source', '') + ':' + str(g.server.id)

    kwargs['headers'] = headers

    cookies = request.cookies

    kwargs['cookies'] = cookies

    return requests.request(request.method, url, verify=verify, **kwargs)


def set_source():
    # not call get_or_raise to allow make requests without D-Source header
    if not hasattr(g, 'source'):
        source_id, proxies = (request.headers.get('D-Source', '') + ':').split(':', 1)
        proxies = proxies.strip(':')
        source = None
        if source_id:
            source = Server.query.get(source_id)
        # check hidden ip on server
        if proxies:
            lp = proxies.split(':')
            neighbour = None
            if lp[-1]:
                neighbour = Server.query.get(lp[-1])
            if neighbour:
                save_if_hidden_ip(request.remote_addr, neighbour)
        if not source:
            source = request.headers.get('D-Source') or request.remote_addr
        elif not proxies:
            save_if_hidden_ip(request.remote_addr, source)

        if not isinstance(source, Server):
            servers = db.session.query(Server).filter_by(deleted=False).join(Gate).filter_by(ip=source).all()
            if len(servers) == 1:
                source = servers[0]

        g.source = source


def forward_or_dispatch(*methods):
    def inner(func):
        from flask import request
        logger = logging.getLogger('dimensigon.forwarder')

        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            destination_id = None
            if 'D-Destination' in request.headers:
                destination_id = request.headers['D-Destination']
            else:
                # Get information from content
                # Code Compatibility. Use D-Destination header instead
                data = request.get_json()
                if data is not None and 'destination' in data:
                    destination_id = data.get('destination')

            if destination_id and destination_id != str(g.server.id) and (not methods or request.method in methods):
                destination: Server = Server.query.get(destination_id)
                if destination is None:
                    return errors.format_error_response(errors.EntityNotFound('Server', destination_id))
                try:
                    logger.debug(f"Forwarding request {request.method} {request.full_path} to {destination.route}")
                    resp = _proxy_request(request=request, destination=destination)
                except requests.exceptions.RequestException as e:
                    return errors.format_error_response(errors.ProxyForwardingError(destination, e))
                else:
                    return resp.content, resp.status_code, dict(resp.headers)

            else:

                value = func(*args, **kwargs)
                return value

        return wrapper_decorator

    return inner


def securizer(func):
    from flask import request
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):

        # cipher_key = session.get('cipher_key', None)
        cipher_key = None
        securizer_method = None

        if 'D-Securizer' in request.headers:
            securizer_method = request.headers.get('D-Securizer')
            if securizer_method == 'plain' and not current_app.config.get('SECURIZER_PLAIN', False):
                    return {'error': 'plain data is not allowed'}, 406

        if request.method != 'GET':
            if request.is_json:
                if securizer_method == 'plain':
                    pass
                else:
                    g.original_json = request.get_json()
                    cipher_key = base64.b64decode(request.json.get('key')) if 'key' in request.json else cipher_key
                    # session['cipher_key'] = cipher_key
                    try:
                        if request.path in url_for('api_1_0.join'):
                            temp_pub_key = rsa.PublicKey.load_pkcs1(request.json.pop('my_pub_key').encode('ascii'))
                            data = unpack_msg2(data=request.get_json(),
                                               pub_key=temp_pub_key,
                                               priv_key=getattr(getattr(g, 'dimension', None), 'private', None),
                                               cipher_key=cipher_key)
                        else:
                            data = unpack_msg(data=request.get_json())
                    except (rsa.pkcs1.VerificationError, NotValidMessage) as e:
                        return {'error': str(e),
                                'message': request.get_json()}, 400
                    if current_app.config.get('SECURIZER', None):
                        # if data and isinstance(data, dict):
                        #     # fill json request with unpacked data
                        #     for key in list(request.json.keys()):
                        #         request.json.pop(key)
                        #     for key, val in data.items():
                        #         request.json[key] = val
                        # else:
                        request._cached_data = json.dumps(data)
                        request._cached_json = (data, data)


            else:
                if request.data:
                    return {'error': 'Content Type must be application/json'}, 400

        rest = tuple()
        rv = func(*args, **kwargs)
        if isinstance(rv, tuple):
            len_rv = len(rv)

            if len_rv >= 2:
                rest = rv[1:]
                rv = rv[0]

        if rv is None:
            return rv

        if isinstance(rv, dict):

            if request.path == url_for('api_1_0.join'):
                rv = pack_msg2(data=rv, pub_key=temp_pub_key,
                               priv_key=getattr(getattr(g, 'dimension', None), 'private', None),
                               cipher_key=cipher_key)
            else:
                if securizer_method == 'plain' and current_app.config.get('SECURIZER_PLAIN', False):
                    pass
                else:
                    rv = pack_msg(data=rv)

        if isinstance(rv, list):
            if securizer_method == 'plain' and current_app.config.get('SECURIZER_PLAIN', False):
                pass
            else:
                rv = pack_msg(data=rv)

        if rest:
            rv = (rv,) + rest
        return rv

    return wrapper_decorator


def validate_schema(schema_name=None, **methods):
    from flask import request
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kw):
            schema = methods.get(request.method.upper()) or methods.get(request.method.lower()) or schema_name
            if schema:
                try:
                    validate(request.json, schema)
                except SchemaError as e:
                    current_app.logger.error(f"Error validating: {json.dumps(request.json, indent=2)}")
                    raise e
            return f(*args, **kw)

        return wrapper

    return decorator


def lock_catalog(f):
    @functools.wraps(f)
    def wrapper(*args, **kw):
        # check if applicant in jwt
        claims = get_jwt_claims()
        if claims:
            if claims.get('applicant'):
                locker = Locker.query.get(Scope.CATALOG)
                if locker.state == State.LOCKED and locker.applicant == claims.get('applicant'):
                    try:
                        ret = f(*args, **kw)
                    except Exception as e:
                        db.session.rollback()
                        raise
                    return ret

        servers = get_servers_from_scope(Scope.CATALOG)
        with lock_scope(Scope.CATALOG, servers):
            try:
                ret = f(*args, **kw)
            except Exception as e:
                db.session.rollback()
                raise
        return ret

    return wrapper


def rollback_on_error(f):
    @functools.wraps(f)
    def wrapper(*args, **kw):
        try:
            return f(*args, **kw)
        except Exception:
            db.session.rollback()
            raise

    return wrapper


def run_as(username: str):
    from flask import Flask
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            # app resolution
            app = None
            if 'app' in kwargs:
                app = kwargs['app']
            else:
                for a in args:
                    if isinstance(a, Flask):
                        app = a
                        break
                else:
                    app = current_app
            if app:
                ctx = app.app_context()
                ctx.push()
            try:
                user = User.get_by_user(username)
                if user is None:
                    raise errors.EntityNotFound("User", username, ['name'])
                from flask_jwt_extended.utils import ctx_stack
                ctx_stack.top.jwt_user = user
                jwt = {'identity': str(user.id)}
                ctx_stack.top.jwt = jwt
                return f(*args, **kwargs)
            finally:
                db.session.close()
                if app:
                    ctx.pop()

        return wrapper

    return decorator


logger = logging.getLogger('dimensigon.time')

def log_time(tag=''):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            start = time.time()
            r = f(*args, **kwargs)
            elapsed = time.time() - start
            logger.debug(
                f"{getattr(f, '__name__', '')} {f'({tag})' if tag else ''} took {elapsed} seconds to complete")
            return r
        return wrapper

    return decorator



import json
import traceback
import typing as t
from datetime import datetime
from http.client import HTTPException

import flask
from flask import Blueprint, jsonify, current_app
from jsonschema import ValidationError
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import InternalServerError

from dimensigon import defaults
from dimensigon.utils.decorators import reify
from dimensigon.utils.helpers import is_iterable_not_string, is_string_types, format_exception
from dimensigon.utils.typos import Id

if t.TYPE_CHECKING:
    from dimensigon.web.network import Response
    from dimensigon.domain.entities import Scope, State, Server
    from dimensigon.domain.entities.route import RouteContainer

bp_errors = Blueprint('errors', __name__)


class BaseError(Exception):
    status_code = 400

    def _format_error_msg(self) -> str:
        pass

    def format(self) -> str:
        return self._format_error_msg()

    @property
    def payload(self) -> t.Optional[dict]:
        return self.__dict__

    def __str__(self):
        msg = self.format()
        payload = None
        if self.payload:
            try:
                payload = json.dumps(self.payload, indent=4)
            except:
                try:
                    payload = str(self.__dict__)
                except:
                    pass
        if msg and payload:
            return f"{msg}\n{payload}"
        else:
            return msg if msg else payload


def format_error_content(error, debug=False):
    content = {
        'error': {
            'type': error.__class__.__name__,
            'message': error.format()
        }
    }

    if error.payload:
        content['error'].update(error.payload)
    if debug:
        content['error'].update(traceback=format_exception(error))
    return content


def format_error_response(error) -> flask.Response:
    content = format_error_content(error)
    rv = jsonify(content)
    rv.status_code = error.status_code or 500

    return rv


@bp_errors.app_errorhandler(BaseError)
def handle_error(error):
    return format_error_response(error)


@bp_errors.app_errorhandler(ValidationError)
def validation_error(error: ValidationError):
    response = {"error": {'type': error.__class__.__name__,
                          'message': error.message,
                          'path': list(error.absolute_path),
                          'schema': error.schema}}
    return response, 400


@bp_errors.app_errorhandler(HTTPException)
def handle_HTTP_exception(e):
    """Return JSON instead of HTML for HTTP errors."""
    # start with the correct headers and status code from the error
    response = e.get_response()
    # replace the body with JSON
    response.data = json.dumps({'error': {'type': e.name, 'message': e.description}})
    response.content_type = "application/json"
    return response


@bp_errors.app_errorhandler(InternalServerError)
def handle_500(error):
    status_code = 500
    original = getattr(error, "original_exception", None)

    response = {'error': {}}

    if current_app.config['DEBUG']:
        tb = [call_string.splitlines() for call_string in traceback.format_tb(original.__traceback__)]
        tb = [val for sublist in tb for val in sublist]
        response['error'].update(type=original.__class__.__name__,
                                 message=str(original),
                                 traceback=tb)
    else:
        response['error'].update(type=error.__class__.__name__, message=error.description)
    return jsonify(response), status_code


class GenericError(BaseError):

    def __init__(self, message, status_code=None, **payload):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self._payload = payload

    def _format_error_msg(self):
        return self.message

    @property
    def payload(self) -> dict:
        return self._payload


class ServerNormalizationError(BaseError):
    status_code = 404

    def __init__(self, idents: t.List[str]):
        self.idents = idents

    def _format_error_msg(self) -> str:
        return f"Servers not found in catalog: {', '.join(self.idents)}"


class UserForbiddenError(BaseError):
    status_code = 404

    def _format_error_msg(self) -> str:
        return "User has no rights to perform the action"


class AlreadyExists(BaseError):
    status_code = 404

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def _format_error_msg(self) -> str:
        return f"already exists in database"

    @property
    def payload(self) -> t.Optional[dict]:
        return {self.name: self.value}

class EntityNotFound(BaseError):
    status_code = 404

    def __init__(self, entity: str, iden, columns: t.List[str] = None):
        self.entity = entity
        self.id = iden
        if is_iterable_not_string(columns):
            self.columns = columns
        elif is_string_types(columns):
            self.columns = [columns]
        else:
            self.columns = ['id']

    def _format_error_msg(self) -> str:
        return f"{self.entity} with given {', '.join(self.columns)} doesn't exist"

    @property
    def payload(self) -> t.Optional[dict]:
        return dict(entity=self.entity, id=self.id)


class NoDataFound(BaseError):
    status_code = 404

    def __init__(self, entity):
        self.entity = entity

    def _format_error_msg(self) -> str:
        return f"No data found"


class UnknownServer(BaseError):
    status_code = 404

    def __init__(self, server_id: str):
        self.server_id = server_id

    def _format_error_msg(self) -> str:
        return f"Unknown Server '{self.server_id}'"


class ServerDeleteError(BaseError):
    status_code = 404

    def _format_error_msg(self) -> str:
        return f"Server to delete must not be de current server"


class ObsoleteCatalog(BaseError):
    status_code = 409

    def __init__(self, actual_catalog: datetime, obsolete_catalog: datetime):
        self.actual_catalog = actual_catalog
        self.obsolete_catalog = obsolete_catalog

    def _format_error_msg(self) -> str:
        return f"Catalog is not up to date"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'current': self.actual_catalog.strftime(defaults.DATEMARK_FORMAT),
                'old': self.obsolete_catalog.strftime(defaults.DATEMARK_FORMAT)}


class CatalogMismatch(BaseError):
    status_code = 500

    def __init__(self, local_entities, remote_entities):
        self.local_entities = local_entities
        self.remote_entities = remote_entities

    def _format_error_msg(self) -> str:
        return "List entities do not match"

#################
# Locker Errors #
#################

class LockerError(BaseError):
    status_code = 409
    action_map = {'L': 'lock', 'U': 'unlock', 'P': 'prevent'}

    def __init__(self, scope):
        self.scope = scope

    @property
    def payload(self) -> t.Optional[dict]:
        return {'scope': self.scope.name}

class NoServerToLock(BaseError):

    def _format_error_msg(self) -> str:
        return 'No server was found for locking'

class PriorityLocker(LockerError):
    def _format_error_msg(self) -> str:
        return 'Priority locker acquired'


class ApplicantLockerError(LockerError):
    def _format_error_msg(self) -> str:
        return f'Another applicant has the scope acquired'


class StatusLockerError(LockerError):

    def __init__(self, scope: 'Scope', action: str, current_state: 'State'):
        super().__init__(scope)
        self.action = self.action_map.get(action, action)
        self.current_state = current_state

    def _format_error_msg(self) -> str:
        return f"Unable to perform locker state transition"

    @property
    def payload(self) -> t.Optional[dict]:
        p = super().payload
        p.update(action=self.action, state=self.current_state.name)
        return p


class LockError(LockerError):

    def __init__(self, scope: 'Scope', action: str, responses: t.List['Response']):
        self.scope = scope
        self.action = self.action_map.get(action, action)
        self.responses = responses

    def _format_error_msg(self) -> str:
        return f"Unable to {self.action} on scope {self.scope.name}"

    @property
    def payload(self) -> t.Optional[dict]:
        responses = []
        for r in self.responses:
            try:
                data = dict({'id': str(r.server.id), 'name': r.server.name})
            except Exception:
                data = dict()
            if r.code:
                data.update(response=r.msg, code=r.code)
            else:
                if isinstance(r.exception, BaseError):
                    data.update(format_error_content(r.exception))
                else:
                    data.update(error=str(r.exception) if str(r.exception) else r.exception.__class__.__name__)
            responses.append(data)
        return {'servers': responses}

    @property
    def status_code(self):
        codes = set([r.code for r in self.responses])
        if len(codes) == 1:
            return codes.pop()
        else:
            if any(map(lambda c: c is None or c >= 500, list(codes))):
                return 500
            else:
                return super().status_code


########################
# Orchestration Errors #
########################
class TargetUnspecified(BaseError):
    status_code = 404

    def __init__(self, target: t.Iterable[str]):
        self.target = list(target)

    def _format_error_msg(self) -> str:
        return f"Target not specified"


class TargetNotNeeded(BaseError):
    status_code = 400

    def __init__(self, target: t.Iterable[str]):
        self.target = list(target)

    def _format_error_msg(self) -> str:
        return f"Target not in orchestration"


class EmptyTarget(BaseError):
    status_code = 400

    def __init__(self, target: str):
        self.target = target

    def _format_error_msg(self) -> str:
        return f"Target does not have any host specified"


class DuplicatedId(BaseError):
    status_code = 400

    def __init__(self, rid):
        self.rid = rid

    def _format_error_msg(self) -> str:
        return "Id already exists"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'id': self.rid}


class UndoStepWithoutParent(BaseError):

    def __init__(self, step_id: Id) -> None:
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "'undo' step must have a parent step"


class ParentUndoError(BaseError):

    def __init__(self, step_id: Id, parent_step_ids: t.Union[Id, t.List[Id]]) -> None:
        if isinstance(parent_step_ids, str):
            self.parent_step_ids = [parent_step_ids]
        else:
            self.parent_step_ids = parent_step_ids
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "a 'do' step cannot have parent 'undo' steps"


class ChildDoError(BaseError):

    def __init__(self, step_id: Id, child_step_ids: t.Union[Id, t.List[Id]]) -> None:
        if isinstance(child_step_ids, str):
            self.child_step_ids = [child_step_ids]
        else:
            self.child_step_ids = child_step_ids
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "an 'undo' step cannot have child 'do' steps"


class CycleError(BaseError):

    def _format_error_msg(self) -> str:
        return "Cycle detected while trying to add dependency"

##############
# Deployment #
##############

class Timeout(BaseError):
    status_code = 500

class RemoteServerTimeout(BaseError):
    status_code = 500

    def __init__(self, timeout, server, command):
        self.timeout = timeout
        self.server = server
        self.command = command

    def _format_error_msg(self) -> str:
        return "Timeout reached waiting remote server operation completion"


#############
# Send File #
#############

class SoftwareServerNotFound(BaseError):
    status_code = 404

    def __init__(self, software_id, server_id):
        self.software_id = software_id
        self.server_id = server_id

    def _format_error_msg(self) -> str:
        return "Software Server Association not found"


class NoSoftwareServer(BaseError):
    status_code = 500

    def __init__(self, software_id):
        self.software_id = software_id

    def _format_error_msg(self) -> str:
        return "Software does not have any server location"


class ChunkSendError(BaseError):
    status_code = 500

    def __init__(self, chunk_responses: t.Dict[int, 'Response']):
        self.chunk_responses = chunk_responses

    def _format_error_msg(self) -> str:
        return "Error while trying to send chunks to server"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'chunks': {c: r.to_dict() for c, r in self.chunk_responses.items()}}


class TransferBase(BaseError):
    status_code = 409


class TransferFileAlreadyExists(TransferBase):

    def __init__(self, file=None):
        self.file = file

    def _format_error_msg(self) -> str:
        return "File already exists. Use force=True if needed"


class TransferFileAlreadyOpen(TransferBase):

    def __init__(self, file=None):
        self.file = file

    def _format_error_msg(self) -> str:
        return "There is already a transfer sending file"


class TransferSoftwareAlreadyOpen(TransferBase):

    def __init__(self, software_id=None):
        self.software_id = software_id

    def _format_error_msg(self) -> str:
        return "There is already a transfer sending software"


class TransferNotInValidState(TransferBase):
    status_code = 410

    def __init__(self, transfer_id: str, status: str):
        self.status = status
        self.transfer_id = transfer_id

    def _format_error_msg(self) -> str:
        return "Transfer not in a valid state"


#################
# Common Errors #
#################
class UnreachableDestination(BaseError):
    status_code = 503

    def __init__(self, server: 'Server', proxy: 'Server' = None):
        from dimensigon.domain.entities import Server
        try:
            p = proxy or Server.get_current()
        except NoResultFound:
            p = None

        self.destination = dict(name=server.name, id=server.id)
        self.proxy = dict(name=p.name, id=p.id) if p else None

    def _format_error_msg(self) -> str:
        return f"Unreachable destination"

    @property
    def payload(self) -> t.Optional[dict]:
        data = dict(destination=self.destination)
        if self.proxy:
            data.update(proxy=self.proxy)
        return data


class KeywordReserved(BaseError):
    status_code = 400

    def __init__(self, msg):
        self.msg = msg

    def _format_error_msg(self) -> str:
        return self.msg


class FileNotFound(BaseError):
    status_code = 404

    def __init__(self, file):
        self.file = file

    def _format_error_msg(self) -> str:
        return "File not found"


class HTTPError(BaseError):

    def __init__(self, resp: 'Response'):
        self.resp = resp
        self.status_code = resp.code if resp.code else 500

    def _format_error_msg(self) -> str:
        return "Error on request"

    @property
    def payload(self) -> t.Optional[dict]:
        return self.resp.to_dict()


class ProxyForwardingError(BaseError):
    status_code = 502

    def __init__(self, dest, exception):
        self.dest = dest
        self.exception = exception

    def _format_error_msg(self) -> str:
        return "Error while trying to forward request"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'server': {'id': str(self.dest.id), 'name': self.dest.name},
                'exception': str(self.exception) if str(self.exception) else self.exception.__class__.__name__}


class HealthCheckMismatch(BaseError):
    status_code = 500

    def __init__(self, expected: t.Dict[str, str], actual: t.Dict[str, str]):
        self.expected = expected
        self.actual = actual

    def _format_error_msg(self) -> str:
        return "Healtcheck response does not match with the server requested"


class ParameterMustBeSet(BaseError):
    status_code = 404

    def __init__(self, msg):
        self.msg = msg

    def _format_error_msg(self) -> str:
        return self.msg


class InvalidRoute(BaseError):
    status_code = 500

    def __init__(self, destination: 'Server', rc: 'RouteContainer'):
        self.destination = destination
        self.rc = rc
        # to load data
        self.payload

    def _format_error_msg(self) -> str:
        return "trying to set an invalid Route"

    @reify
    def payload(self) -> t.Optional[dict]:
        data = {'route': {'destination': {'id': self.destination.id, 'name': self.destination.name},
                          'proxy_server': None, 'gate': None, 'cost': None}}
        if self.rc.proxy_server:
            data['route']['proxy_server'] = {'id': self.rc.proxy_server.id, 'name': self.rc.proxy_server.name}
        if self.rc.gate:
            data['route']['gate'] = {'id': self.rc.gate.id, 'gate': str(self.rc.gate)}
        if self.rc.cost is not None:
            data['route']['cost'] = self.rc.cost

        return data

class InvalidDateFormat(BaseError):

    def __init__(self, date: str, expected_format: str):
        self.input_date = date
        self.expected_format = expected_format

    def _format_error_msg(self) -> str:
        return "Date is not a valid format"


import responses
from aioresponses import aioresponses

from dimensigon.domain.entities import Dimension, Server, Route, Gate
from dimensigon.utils.helpers import get_now
from dimensigon.web import create_app, db
from dimensigon.web.network import ping
from tests.base import TestDimensigonBase
from tests.helpers import generate_dimension_json_data, set_callbacks

now = get_now()

class TestPing(TestDimensigonBase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.dim = generate_dimension_json_data()

    def fill_database(self, node):
        db.create_all()
        d = Dimension.from_json(self.dim)
        s1 = Server(id='00000000-0000-0000-0000-000000000001', name='node1', created_on=now, me=node == 'node1')
        g11 = Gate(id='00000000-0000-0000-0000-000000000011', server=s1, port=5000, dns=s1.name)
        s2 = Server(id='00000000-0000-0000-0000-000000000002', name='node2', created_on=now, me=node == 'node2')
        g12 = Gate(id='00000000-0000-0000-0000-000000000012', server=s2, port=5000, dns=s2.name)
        s3 = Server(id='00000000-0000-0000-0000-000000000003', name='node3', created_on=now, me=node == 'node3')
        g13 = Gate(id='00000000-0000-0000-0000-000000000013', server=s3, port=5000, dns=s3.name)
        s4 = Server(id='00000000-0000-0000-0000-000000000004', name='node4', created_on=now, me=node == 'node4')
        g14 = Gate(id='00000000-0000-0000-0000-000000000014', server=s4, port=5000, dns=s4.name)

        if node == 'node1':
            self.s1 = s1
            self.s2 = s2
            self.s3 = s3
            self.s4 = s4
            Route(s2, g12)
            Route(s3, s2, 1)
        elif node == 'node2':
            Route(s1, g11)
            Route(s3, g13)
        elif node == 'node3':
            Route(s1, s2, 1)
            Route(s2, g12)

        db.session.add_all([d, s1, s2, s3, s4])
        db.session.commit()

    def setUp(self) -> None:
        self.maxDiff = None
        super().setUp()
        self.app2 = create_app('test')
        self.app2.name = 'dimensigon2'
        self.client2 = self.app2.test_client()
        self.app3 = create_app('test')
        self.app3.name = 'dimensigon3'
        self.client3 = self.app3.test_client()

        # fill data
        self.fill_database('node1')

        with self.app2.app_context():
            self.fill_database('node2')

        with self.app3.app_context():
            self.fill_database('node3')

    @aioresponses()
    @responses.activate
    def test_ping(self, m):
        set_callbacks([('node1', self.client), ('node2', self.client2), ('node3', self.client3)], m)

        cost, time = ping(self.s1)
        self.assertEqual(0, cost)

        cost, time = ping(self.s2)
        self.assertEqual(0, cost)

        cost, time = ping(self.s3)
        self.assertEqual(1, cost)

        cost, time = ping(self.s4)
        self.assertEqual(None, cost)



import datetime as dt
import typing as t

from dimensigon.utils.typos import Dill
from dimensigon.web import db


class _Empty:
    pass


class Parameter(db.Model):
    __tablename__ = 'L_parameter'

    parameter = db.Column(db.String(80), primary_key=True)
    value = db.Column(db.Text)
    dump = db.Column(Dill)
    load = db.Column(Dill)

    def __init__(self, parameter: str, dump: t.Callable[[t.Any], str] = None, load: t.Callable[[str], t.Any] = None):
        self.parameter = parameter
        self.dump = dump
        self.load = load

    @staticmethod
    def _convert(param: 'Parameter'):
        if param.load:
            try:
                value = param.load(param.value)
            except:
                return None
            else:
                return value
        else:
            return param.value

    @classmethod
    def _get_parameter(cls, key: str):
        p = cls.query.get(key)
        if p:
            return p
        else:
            raise KeyError(key)

    @classmethod
    def get(cls, key: str, default=_Empty):
        try:
            p = cls._get_parameter(key)
        except KeyError:
            if default == _Empty:
                raise
            else:
                return default
        value = cls._convert(p)
        if value is None and default != _Empty:
            return default
        else:
            return value

    def set_functions(self, dump: t.Callable[[t.Any], str] = _Empty, load: t.Callable[[str], t.Any] = _Empty):
        if dump != _Empty:
            self.dump = dump
        if load != _Empty:
            self.load = load

    @classmethod
    def set(cls, key, value):
        p = cls.query.get(key)
        if p.dump:
            p.value = p.dump(value)
        else:
            p.value = value

    @classmethod
    def set_initial(cls, session=None):
        if session is None:
            session = db.session

        if not session.query(cls).get('last_graceful_shutdown'):
            p = Parameter('last_graceful_shutdown', dump=lambda x: x.strftime("%m/%d/%Y, %H:%M:%S.%f %z"),
                          load=lambda x: dt.datetime.strptime(x, "%m/%d/%Y, %H:%M:%S.%f %z"))
            session.add(p)
        if not session.query(cls).get('fetching_catalog'):
            p = Parameter('fetching_catalog', dump=lambda x: '1' if x else '0',
                          load=lambda x: True if x == '1' else False)
            session.add(p)
        if not session.query(cls).get('join_server'):
            p = Parameter('join_server')
            session.add(p)


import logging
import os


basedir = os.path.abspath(os.path.dirname(__file__))


class Config(object):
    CONFIG_FOLDER = None
    DEBUG = False
    TESTING = False
    CSRF_ENABLED = True
    SECRET_KEY = os.environ.get('DM_SECRET_KEY') or 'hard to guess string'
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    DATETIME_FORMAT = "%d/%m/%Y, %H:%M:%S"
    SSL_REDIRECT = False
    SSL_VERIFY = False
    PROPAGATE_EXCEPTIONS = False

    # executor
    EXECUTOR_MAX_WORKERS = min(32, os.cpu_count() + 4)
    EXECUTOR_PROPAGATE_EXCEPTIONS = True

    AUTOUPGRADE = True
    PREFERRED_URL_SCHEME = 'https'  # scheme used to communicate with servers
    SECURIZER = True
    SECURIZER_PLAIN = True
    SCHEDULER = True

    @classmethod
    def init_app(cls, app):
        pass


class ProductionConfig(Config):
    SQLALCHEMY_DATABASE_URI = None

    @classmethod
    def init_app(cls, app):
        Config.init_app(app)

        # email errors to the administrators
        # import logging
        # from logging.handlers import SMTPHandler
        # credentials = None
        # secure = None
        # if getattr(cls, 'MAIL_USERNAME', None) is not None:
        #     credentials = (cls.MAIL_USERNAME, cls.MAIL_PASSWORD)
        #     if getattr(cls, 'MAIL_USE_TLS', None):
        #         secure = ()
        # mail_handler = SMTPHandler(
        #     mailhost=(cls.MAIL_SERVER, cls.MAIL_PORT),
        #     fromaddr=cls.FLASKY_MAIL_SENDER,
        #     toaddrs=[cls.FLASKY_ADMIN],
        #     subject=cls.FLASKY_MAIL_SUBJECT_PREFIX + ' Application Error',
        #     credentials=credentials,
        #     secure=secure)
        # mail_handler.setLevel(logging.ERROR)
        # app.logger.addHandler(mail_handler)

        # import logging
        # from logging import StreamHandler
        # stream_handler = StreamHandler()
        # stream_handler.setLevel(logging.INFO)
        # app.logger.addHandler(stream_handler)


class GunicornConfig(ProductionConfig):

    @classmethod
    def init_app(cls, app):
        Config.init_app(app)
        # from logging import FileHandler
        # file_handler = FileHandler('dimensigon.log')
        # file_handler.setLevel(logging.INFO)
        # app.logger.addHandler(file_handler)
        # fmt = logging.Formatter(
        #     "%(asctime)s [%(process)d] [%(module)s] [%(funcName)s] [%(name)s] [%(levelname)s] %(message)s")
        # file_handler.setFormatter(fmt)
        # for hdlr in app.logger.handlers:
        #     hdlr.setFormatter(fmt)


class UnixConfig(ProductionConfig):
    @classmethod
    def init_app(cls, app):
        ProductionConfig.init_app(app)

        # log to syslog
        import logging
        from logging.handlers import SysLogHandler
        syslog_handler = SysLogHandler()
        syslog_handler.setLevel(logging.WARNING)
        app.logger.addHandler(syslog_handler)


class TestingConfig(Config):
    TESTING = True
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    AUTOUPGRADE = False
    SERVER_NAME = 'test'
    PREFERRED_URL_SCHEME = 'http'
    SECURIZER = False
    DEBUG = False

    @classmethod
    def init_app(cls, app):
        super().init_app(app)
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for logger in loggers:
            logger.handlers = []
        logging.root.handlers = []


class DevelopmentConfig(Config):
    DEVELOPMENT = True
    DEBUG = True
    ENV = 'development'
    SQLALCHEMY_DATABASE_URI = os.environ.get('DM_DEV_DATABASE_URL') or \
                              'sqlite:///' + os.path.join(basedir, 'dimensigon-dev.db')
    AUTOUPGRADE = False

    @classmethod
    def init_app(cls, app):
        Config.init_app(app)
        import logging

        for h in app.logger.handlers:
            h.setLevel(logging.DEBUG)


config_by_name = dict(
    development=DevelopmentConfig(),
    test=TestingConfig(),
    production=ProductionConfig(),
    default=ProductionConfig(),
    unix=UnixConfig(),
    gunicorn=GunicornConfig()
)



import datetime as dt
import hashlib
import importlib
import inspect
import itertools
import logging
import os
import re
import sys
import traceback
import typing as t
from collections import Iterable
from contextlib import contextmanager

import netifaces
import six
from cryptography.fernet import Fernet
from flask import current_app

from dimensigon import defaults

_LOGGER = logging.getLogger(__name__)


class Singleton(type):
    """ Metaclass that creates a Singleton base type when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls)\
                .__call__(*args, **kwargs)
        return cls._instances[cls]


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def is_string_types(var):
    return isinstance(var, six.string_types)


def is_iterable(var):
    return isinstance(var, Iterable)


def is_iterable_not_string(var):
    return is_iterable(var) and not is_string_types(var)


UUID_PATTERN = re.compile(r'^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$')


def is_valid_uuid(var):
    # best performance with regular expression compiled and not using re.IGNORECASE
    if UUID_PATTERN.match(var):
        return True
    else:
        return False


def convert(d):
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = convert(v)
    return AttributeDict(d)


# function to mock datetime.now
def get_now() -> dt.datetime:
    return dt.datetime.now(dt.timezone.utc)


def str_to_key(id_: str):
    key = []
    for x in id_.split('.'):
        try:
            x = int(x)
        except ValueError:
            pass
        finally:
            key.append(x)
    return tuple(key) if len(key) > 1 else key[0]


def key_to_str(key):
    if isinstance(key, t.Sequence):
        id_ = '.'.join([str(i) for i in key])
    else:
        id_ = str(key)
    return id_


def generate_url(destination, uri, protocol='https'):
    from dimensigon.domain.entities import Server
    try:
        forwarder = destination.mesh_best_route[0]
    except IndexError:
        forwarder = destination
    else:
        forwarder = Server.query.get(forwarder)

    return f"{protocol}://{forwarder.name}:{forwarder.port}{uri}"


def generate_symmetric_key():
    return Fernet.generate_key()


def encrypt_symmetric(data, key):
    cipher_suite = Fernet(key)
    return cipher_suite.encrypt(data)


def decrypt_symmetric(data, key):
    cipher_suite = Fernet(key)
    return cipher_suite.decrypt(data)


def encrypt(data: bytes, symmetric_key: bytes = None) -> \
        t.Tuple[bytes, t.Optional[bytes]]:
    """

    Parameters
    ----------
    data:
        data to encrypt.
    symmetric_key:
        symmetric key used for encrypting data. If set, cipher_key must be None

    Returns
    -------

    """
    new_symmetric_key = None
    if not symmetric_key:
        symmetric_key = new_symmetric_key = generate_symmetric_key()
    cipher_data = encrypt_symmetric(data, symmetric_key)
    return cipher_data, new_symmetric_key


def decrypt(cipher_text: bytes, symmetric_key: bytes) -> bytes:
    """

    Parameters
    ----------
    cipher_text:
        text to decrypt
    cipher_key:
        symmetric_key encrypted. If specified symmetric_key must be None (default)
    symmetric_key:
        symmetric_key. If specified cipher_token must be None (default)
    key:
        key used for decryption of cipher_key

    Returns
    -------
    decrypted data
    """
    cipher_suite = Fernet(symmetric_key)
    dumped_data = cipher_suite.decrypt(cipher_text)
    return dumped_data


def get_logger(self=None):
    if self:
        name = '.'.join([
            self.__module__,
            self.__name__
        ])
    else:
        name = None
    try:
        current_app.logger
    except RuntimeError:
        logger = logging.getLogger(name)
    else:
        logger = current_app.logger
    return logger


def get_entities() -> t.List[t.Tuple['str', t.Any]]:
    from dimensigon.web import db
    entities = []
    for name, cls in inspect.getmembers(sys.modules['dimensigon.domain.entities'],
                                        lambda x: (inspect.isclass(x) and issubclass(x, db.Model))):
        entities.append((name, cls))

    return entities


def get_distributed_entities() -> t.List[t.Tuple['str', t.Any]]:
    from dimensigon.domain.entities.base import DistributedEntityMixin
    entities = []
    for name, cls in inspect.getmembers(sys.modules['dimensigon.domain.entities'],
                                        lambda x: (inspect.isclass(x) and issubclass(x, DistributedEntityMixin))):
        entities.append((name, cls))

    return sorted(entities, key=lambda x: x[1].order or 99999)


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_filename_from_cd(cd):
    """
    Get filename from content-disposition
    """
    if not cd:
        return None
    fname = re.findall('filename=(.+)', cd)
    if len(fname) == 0:
        return None
    return fname[0]


def generate_dimension(name):
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from dimensigon.domain.entities import Dimension

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend()
    )
    priv_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.TraditionalOpenSSL,
                                         encryption_algorithm=serialization.NoEncryption())
    pub_pem = private_key.public_key().public_bytes(encoding=serialization.Encoding.PEM,
                                                    format=serialization.PublicFormat.PKCS1)

    return Dimension(name=name, private=priv_pem, public=pub_pem)


def remove_prefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


def get_config_from_filename(filename):
    if not os.path.exists(filename):
        raise RuntimeError("%r doesn't exist" % filename)

    ext = os.path.splitext(filename)[1]

    try:
        module_name = '__config__'
        if ext in [".py", ".pyc"]:
            spec = importlib.util.spec_from_file_location(module_name, filename)
        else:
            msg = "configuration file should have a valid Python extension.\n"
            loader_ = importlib.machinery.SourceFileLoader(module_name, filename)
            spec = importlib.util.spec_from_file_location(module_name, filename, loader=loader_)
        mod = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = mod
        spec.loader.exec_module(mod)
    except Exception:
        print("Failed to read config file: %s" % filename, file=sys.stderr)
        traceback.print_exc()
        sys.stderr.flush()
        sys.exit(1)

    return vars(mod)


SQLITE_URL_PREFIX = "sqlite://"


def validate_sqlite_database(dbpath: str) -> bool:
    """Run a quick check on an sqlite database to see if it is corrupt."""
    import sqlite3  # pylint: disable=import-outside-toplevel

    dbpath = dbpath[len(SQLITE_URL_PREFIX):]

    if not os.path.exists(dbpath):
        # Database does not exist yet, this is OK
        return True

    try:
        conn = sqlite3.connect(dbpath)
        conn.cursor().execute("PRAGMA QUICK_CHECK")
        conn.close()
    except sqlite3.DatabaseError:
        # _LOGGER.exception("The database at %s is corrupt or malformed.", dbpath)
        return False

    return True


def get_ips(ipv4=True, ipv6=False) -> t.List[t.Tuple[str, int]]:
    ips = []

    if ipv4:
        ips.extend(list(itertools.chain(
            *[[ip['addr'] for ip in netifaces.ifaddresses(iface).get(netifaces.AF_INET, [])] for iface in
              netifaces.interfaces()])))
    if ipv6:
        iter_ips = itertools.chain(
            *[[ip['addr'] for ip in netifaces.ifaddresses(iface).get(netifaces.AF_INET6, [])] for iface in
              netifaces.interfaces()])

        for ip in iter_ips:
            if ip:
                if '%' in ip:
                    ips.append(ip.rsplit('%')[0])
                else:
                    ips.append(ip)
    return ips

def bind2gate(bind: t.List[str]) -> t.List[t.Tuple[str, int]]:
    specified_gates = set()
    for b in bind:
        if ':' in b:
            gate = b.rsplit(':')
            gate = (gate[0], int(gate[1]))
        else:
            gate = (b, defaults.DEFAULT_PORT)
        if gate[0] == '0.0.0.0':
            ips = get_ips(ipv4=True, ipv6=False)
            for ip in ips:
                specified_gates.update([(ip, int(gate[1]))])
        elif gate[0] == '::':
            ips = get_ips(ipv4=False, ipv6=True)
            for ip in ips:
                specified_gates.update([(ip, int(gate[1]))])
        else:
            specified_gates.update([gate])
    return list(specified_gates)

def clean_string(incoming_string):
    replace_char = '_'
    newstring = incoming_string
    newstring = newstring.replace("!", replace_char)
    newstring = newstring.replace("@", replace_char)
    newstring = newstring.replace("#", replace_char)
    newstring = newstring.replace("$", replace_char)
    newstring = newstring.replace("%", replace_char)
    newstring = newstring.replace("^", replace_char)
    newstring = newstring.replace("&", replace_char)
    newstring = newstring.replace("*", replace_char)
    newstring = newstring.replace("(", replace_char)
    newstring = newstring.replace(")", replace_char)
    newstring = newstring.replace("+", replace_char)
    newstring = newstring.replace("=", replace_char)
    newstring = newstring.replace("?", replace_char)
    newstring = newstring.replace("\'", replace_char)
    newstring = newstring.replace("\"", replace_char)
    newstring = newstring.replace("{", replace_char)
    newstring = newstring.replace("}", replace_char)
    newstring = newstring.replace("[", replace_char)
    newstring = newstring.replace("]", replace_char)
    newstring = newstring.replace("<", replace_char)
    newstring = newstring.replace(">", replace_char)
    newstring = newstring.replace("~", replace_char)
    newstring = newstring.replace("`", replace_char)
    newstring = newstring.replace(":", replace_char)
    newstring = newstring.replace(";", replace_char)
    newstring = newstring.replace("|", replace_char)
    newstring = newstring.replace("\\", replace_char)
    newstring = newstring.replace("/", replace_char)
    newstring = newstring.replace("-", replace_char)
    return newstring


@contextmanager
def session_scope(session=None):
    """Provide a transactional scope around a series of operations."""
    if session is None:
        raise RuntimeError("Session required")

    need_rollback = False
    try:
        yield session
        if session.transaction:
            need_rollback = True
            session.commit()
    except Exception as err:
        _LOGGER.error("Error executing query: %s", err)
        if need_rollback:
            session.rollback()
        raise
    finally:
        session.close()


def format_exception(exc: Exception):
    return ''.join(traceback.format_exception(exc, exc, exc.__traceback__))



import threading
from unittest import mock
from unittest.case import TestCase

import flask
from flask_jwt_extended import create_access_token

from dimensigon.domain.entities import User
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.web import errors, create_app, db
from tests.helpers import set_test_scoped_session


def app_scope():
    try:
        return str(hash(flask._app_ctx_stack.top.app)) + str(threading.get_ident())
    except:
        return str(threading.get_ident())


class TestCaseLockBypass(TestCase):

    def run(self, result=None):
        with mock.patch('dimensigon.use_cases.lock.lock'):
            with mock.patch('dimensigon.use_cases.lock.unlock'):
                super().run(result)


class ValidateResponseMixin:

    def validate_error_response(self, resp: flask.Response, error: errors.BaseError):
        self.assertEqual(error.status_code, resp.status_code)
        self.assertDictEqual(errors.format_error_content(error), resp.get_json())


class TestDimensigonBase(TestCase, ValidateResponseMixin):

    def set_scoped_session(self, func=app_scope):
        set_test_scoped_session(db, func)

    def setUp(self) -> None:
        self.maxDiff = None
        self.set_scoped_session()
        self.app = create_app('test')
        self.app.config['SECURIZER'] = False
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        User.set_initial()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()



import datetime as dt
import logging
import os
import sys
import typing as t

from gunicorn.app.base import Application

from dimensigon import defaults
from dimensigon.exceptions import DimensigonError
from dimensigon.web import DimensigonFlask, create_app, threading

_LOGGER = logging.getLogger(__name__)


class GunicornApp(Application):
    dm = None

    def __init__(self, application, options=None):
        """ Construct the Application. Default gUnicorn configuration is loaded """

        self.application = application
        self.options = options or {}

        # if port, or host isn't set-- run from os.environments
        #
        super(GunicornApp, self).__init__()

    def init(self, parser, opts, args):
        pass

    def load_config(self):
        # Load up the any app specific configuration
        for k, v in self.options.items():
            self.cfg.set(k.lower(), v)

    def load(self):
        return self.application


class Dimensigon:
    def __init__(self):
        self.flask_app: t.Optional[DimensigonFlask] = None
        self.http_server: t.Optional[GunicornApp] = None
        self.config = Config(self)
        self.engine = None
        self.get_session = None

    def create_flask_instance(self):
        if self.flask_app is None:
            self.flask_app = create_app(self.config.flask_conf)
            self.flask_app.dm = self

    def create_gunicorn_instance(self):
        if self.http_server is None:
            self.http_server = GunicornApp(self.flask_app, self.config.http_conf)
            self.http_server.dm = self

    def create_instances(self):
        self.create_flask_instance()
        self.create_gunicorn_instance()

    def start(self):
        """starts dimensigon server"""
        self.create_instances()
        self.flask_app.bootstrap_start()
        th = threading.Timer(interval=4, function=self.flask_app.make_first_request)
        th.start()
        if self.config.flask:
            self.flask_app.run(host='0.0.0.0', port=defaults.DEFAULT_PORT, ssl_context='adhoc')
            self.flask_app.shutdown()
            sys.exit(0)
        else:
            self.http_server.run()

class Config:

    def __init__(self, dm: Dimensigon):
        self.dm = dm

        # Directory that holds the configuration
        self.config_dir: t.Optional[str] = None

        # If set, process should upgrade as soon as a neighbour has a higher version
        self.auto_upgrade: bool = True

        # sets a security layer on top of HTTP
        self.security_layer: bool = True

        # allow pass through security layer without encrypt packet with header D-Securizer: plain
        self.security_layer_antidote: bool = False

        # runs the scheduler
        self.scheduler: bool = True

        # database uri
        self.db_uri: t.Optional[str] = None

        # http configuration
        self.http_conf = {}

        # flask configuration
        self.flask_conf = {}

        # Run configuration (used for elevator to load same configuration)
        self.run_config = {}

        self.debug: bool = False

        self.flask: bool = False

        # forces the process to scan on startup
        self.force_scan: bool = False

        # Run route table, catalog and cluster refresh every minutes
        self.refresh_interval: dt.timedelta = defaults.REFRESH_PERIOD

    def path(self, *path: str) -> str:
        """Generate path to the file within the configuration directory.
        """
        if self.config_dir is None:
            raise DimensigonError("config_dir is not set")
        return os.path.join(self.config_dir, *path)


from unittest import TestCase

from flask_jwt_extended import create_access_token

from dimensigon import defaults
from dimensigon.domain.entities import Server, File, FileServerAssociation
from dimensigon.web import create_app, db, errors


class TestApi(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        # set_initial(server=False)

        self.srv1 = Server('node1', id='00000000-0000-0000-0000-000000000001', me=True)
        self.file = File(source_server=self.srv1, target='/etc/ssh/sshd_config', id='00000000-0000-0000-0000-000000000002')
        self.fsa = FileServerAssociation(file=self.file, destination_server=self.srv1)

        db.session.add_all([self.srv1, self.file, self.fsa])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_from_json(self):
        fsa_json = self.fsa.to_json()
        fsa_json['dest_folder'] = '/new_target'

        smashed = FileServerAssociation.from_json(fsa_json)

        self.assertEqual(self.fsa.file, smashed.file)
        self.assertIsNotNone(smashed.file)
        self.assertEqual("/new_target", smashed.dest_folder)
        self.assertEqual(self.srv1, smashed.destination_server)
        self.assertIsNotNone(smashed.destination_server)
        self.assertEqual(self.fsa.last_modified_at, smashed.last_modified_at)
        self.assertIsNotNone(smashed.last_modified_at)

        db.session.commit()
        db.session.close()

        del smashed

        fsa = FileServerAssociation.query.get(('00000000-0000-0000-0000-000000000002', self.srv1.id))
        self.assertEqual("/new_target", fsa.dest_folder)

    def test_from_json_new(self):
        fsa_json = dict(file_id=self.file.id,
                        dst_server_id=self.srv1.id,
                        dest_folder='/root',
                        last_modified_at=defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT))

        smashed = FileServerAssociation.from_json(fsa_json)

        self.assertEqual(self.fsa.file, smashed.file)
        self.assertIsNotNone(smashed.file)
        self.assertEqual("/root", smashed.dest_folder)
        self.assertEqual(self.srv1, smashed.destination_server)
        self.assertIsNotNone(smashed.destination_server)
        self.assertEqual(defaults.INITIAL_DATEMARK, smashed.last_modified_at)
        self.assertIsNotNone(smashed.last_modified_at)

        with self.assertRaises(errors.EntityNotFound):
            fsa_json = dict(file_id='unknown',
                            dst_server_id=self.srv1.id,
                            dest_folder='/root',
                            last_modified_at=defaults.INITIAL_DATEMARK.strftime(defaults.DATEMARK_FORMAT))

            FileServerAssociation.from_json(fsa_json)